/*
Copyright (C) 2006 Evan Teran
                   eteran@alum.rit.edu

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include <QApplication>
#include <QDir>
#include <QPluginLoader>
#include <QString>
#include <QMessageBox>
#include <QVector>

#include "DebuggerCoreInterface.h"
#include "DebuggerPluginInterface.h"
#include "DebuggerMain.h"
#include "Debugger.h"
#include "Configuration.h"
#include "version.h"

#include <iostream>
#include <ostream>

static QVector<DebuggerPluginInterface *> generalPlugins;

//------------------------------------------------------------------------------
// Name: loadPlugins(const QString &directory)
// Desc: attemps to load all plugins in a given directory
//------------------------------------------------------------------------------
void loadPlugins(const QString &directory) {
	QDir pluginsDir(qApp->applicationDirPath());
	
	// TODO: attempt to detect the same plugin being loaded twice
	
	pluginsDir.cd(directory);

	foreach (QString fileName, pluginsDir.entryList(QDir::Files)) {
	
		QPluginLoader loader(pluginsDir.absoluteFilePath(fileName));
		
		QObject *const plugin = loader.instance();
		
		if (plugin) {
		
			Debugger::log(QString().sprintf("Loading Plugin: %-30s : %p", qPrintable(fileName), plugin));
			
			// TODO: handle the case where we find more than one core plugin...
			DebuggerCoreInterface *const core_plugin = qobject_cast<DebuggerCoreInterface *>(plugin);
			if(core_plugin != 0) {
				Debugger::debuggerBase = core_plugin;
			}
			
			DebuggerPluginInterface *const generic_plugin = qobject_cast<DebuggerPluginInterface *>(plugin);
			if(generic_plugin != 0) {
				// we have a general purpose plugin
				generalPlugins.push_back(generic_plugin);
			}
		}
	}
}

//------------------------------------------------------------------------------
// Name: main(int argc, char *argv[])
// Desc: entry point
//------------------------------------------------------------------------------
int main(int argc, char *argv[]) {

    QApplication app(argc, argv);
	
	std::vector<std::string> args(argv, argv + argc);
	
	BaseTypes::pid_t attach_pid = 0;
	
	// setup organization info so settings go in right place
	QApplication::setOrganizationName("codef00.com");
	QApplication::setOrganizationDomain("codef00.com");
	QApplication::setApplicationName("EDB");
	
	if(args.size() > 1) {
		if(args.size() == 3) {
			if(args[1] == "--attach") {
				attach_pid = strtoul(args[2].c_str(), 0, 10);
			} else {
				std::cerr << "usage: " << args[0] << " [ --attach <pid> ]" << std::endl;
				return -1;
			}
		} else {
			std::cerr << "usage: " << args[0] << " [ --attach <pid> ]" << std::endl;
			return -1;
		}
	}

	Debugger::log(QString().sprintf("Starting EDB Version: %s", version_string));

	// create the main windows object	
	DebuggerMain mainWin;
	
	// look for some plugins..
	loadPlugins(Debugger::config()->pluginPath);
	
	// let the plugins setup their menus
	mainWin.finishPluginSetup(generalPlugins);
	
	// ok things are initialized to a reasonable degree, let's show the main window
	mainWin.show();
	
	if(attach_pid != 0) {
		mainWin.doAttach(attach_pid);
	}
		
	if(Debugger::debuggerBase == 0) {
		QMessageBox::warning(0, 
			"EDB Failed To Load Neccessary Plugin", 
			"Failed to load debugger core plugin. Please make sure it exists and that the plugin path is correctly configured.\n"
			"This is normal if EDB has not been previously run or the configuration file has been removed.");
		Debugger::dialogOptions()->exec();
		
		// TODO: detect if they corrected the issue and try again
		return -1;
	} else {
	    return app.exec();
	}
}
